/*
 * $Id: vtecapture.c 590 2014-03-06 19:30:23Z elkner $
 *
 * This file and its contents are supplied under the terms of the
 * Common Development and Distribution License ("CDDL"), version 1.0.
 * You may only use this file in accordance with the terms of version
 * 1.0 of the CDDL.
 *
 * A full copy of the text of the CDDL should have accompanied this
 * source.  A copy of the CDDL is also available via the Internet at
 * http://www.illumos.org/license/CDDL.
 */

/*
 * Copyright 2014 Jens Elkner.  All rights reserved.
 */

#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include <glib-object.h>
#include <glib/gstdio.h>
#include <glib/gi18n.h>
#include <vte/vte.h>
#include "vte-private.h"

static void
destroy_and_quit(VteTerminal *terminal, GtkWidget *window)
{
	char *cfname = g_object_get_data (G_OBJECT (window), "cfile");
	int fd = GPOINTER_TO_INT(g_object_get_data (G_OBJECT (window), "fd"));
	if (fd != -1) {
		if (close(fd) == -1) {
			if (close(fd) == -1) {
				g_printerr("Unable to close file '%s': %s\n",
					cfname == NULL ? "???" : cfname, g_strerror (errno));
			}
		} else {
			fd = -1;
		}
	}

	gtk_widget_destroy (window);
	gtk_main_quit ();
}
static void
delete_event(GtkWidget *window, GdkEvent *event, gpointer terminal)
{
	destroy_and_quit(VTE_TERMINAL (terminal), window);
}
static void
child_exited(GtkWidget *terminal, gpointer window)
{
	destroy_and_quit(VTE_TERMINAL (terminal), GTK_WIDGET (window));
}


static void
append_unistr(VteTerminal *terminal, glong row, glong col, int len, GString *gs)
{
	/* f..., this is an index into the mb char table 'unistr_decomp'.
	 * Since it is vte lib private, there is no trivial way to retrieve it.
	 * So try a workaround by getting the text with attrs ... */
	char *text;
	GArray *attrs;
	VteCharAttributes a;
	glong i, last;

	attrs = g_array_new (FALSE, TRUE, sizeof (VteCharAttributes));
	text = vte_terminal_get_text_include_trailing_spaces(terminal, NULL,
			NULL, attrs);
	
	if (text == NULL)
		return;
	/* we know: strlen(text) == #attrs */
	last = G_N_ELEMENTS(attrs);

	/* can't use gs->len because we use a different aproach wrt. EOL and
	 * soft-wrapping */
	for (i=0; i < last; i++) {
		a = g_array_index(attrs, VteCharAttributes, i);
		if (a.row == row) {
			break;
		}
	}
	for (; i < last; i++) {
		a = g_array_index(attrs, VteCharAttributes, i);
		if (a.column == col)
			break;
	}
	if (i < last) {
		col = i + len;
		for (; i < col; i++) {
			g_string_append_unichar(gs, text[i]);
		}
	}
	/* we don't keep it, since we expect this is an exception */
	g_array_free(attrs, TRUE);
	g_free (text);
}


#define STATE_MASK 0xFFFFFFE0

static void
adjust_state(VteIntCellAttr *old, VteIntCellAttr cur, gboolean first,
	GString *gs)
{
	guint32 ci;

	if (old->i == (cur.i & STATE_MASK) && !first)
		return;

	/*	Since it is not clear, whether it is allowed to feed several CSI
		sequences at once, we separate them (even if it would work for some
		terminals)
		See also: vte_terminal_determine_colors_internal()
			vte_sequence_handler_character_attributes()
	 */
	if (first || (old->s.bold ^ cur.s.bold) || (old->s.half ^ cur.s.half)) {
		/* half and bold are mutually exclusive */
		if (cur.s.bold) {
			g_string_append(gs, "\033[1m");
		} else if (cur.s.half) {
			g_string_append(gs, "\033[2m");
		} else if (!first) {
			g_string_append(gs, "\033[22m"); 
		}
	}
	if (first || (old->s.underline ^ cur.s.underline)) {
		if (cur.s.underline) {
			g_string_append(gs, "\033[4m");
		} else if (!first) {
			g_string_append(gs, "\033[24m");
		}
	}
	if (first || (old->s.blink ^ cur.s.blink)) {
		if (cur.s.blink) {
			g_string_append(gs, "\033[5m");
		} else if (!first) {
			g_string_append(gs, "\033[25m");
		}
	}
	/* There is no sequence for text selection. So we need to emulate it by
	   swapping fg and bg, which is the default behavior, if no bg highlite
	   color has been explicitly set. */
	gboolean oreverse = old->s.reverse , creverse = cur.s.reverse;
	if (old->s.selected)
		oreverse = !oreverse;
	if (cur.s.selected)
		creverse = !creverse;
	if (first || (oreverse ^ creverse)) {
		if (creverse) {
			g_string_append(gs, "\033[7m");
		} else if (!first) {
			g_string_append(gs, "\033[27m");
		}
	}
	if (first || (old->s.invisible ^ cur.s.invisible)) {
		if (cur.s.reverse) {
			g_string_append(gs, "\033[8m");
		} else if (!first) {
			g_string_append(gs, "\033[28m");
		}
	}
	if (first || (old->s.strikethrough ^ cur.s.strikethrough)) {
		if (cur.s.strikethrough) {
			g_string_append(gs, "\033[9m");
		} else if (!first) {
			g_string_append(gs, "\033[29m");
		}
	}
	ci = cur.s.fore;
	if ((first && (ci != VTE_DEF_FG)) || (old->s.fore != ci)) {
		if (ci < 8) {
			/* 30..37  base FG */
			g_string_append_printf(gs, "\033[%dm", 30 + ci);
		} else if (ci < VTE_COLOR_DIM_OFFSET) {
			/* 90..97  16-color FG  */
			g_string_append_printf(gs, "\033[%dm",
				90 + ci - VTE_COLOR_BRIGHT_OFFSET);
		} else if (ci == VTE_DEF_FG) {
			/* default FG */
			g_string_append(gs, "\033[39m" );
		} else {
			/* 0..256  256-color FG  */
			g_string_append_printf(gs, "\033[38;5;%dm", ci);
		}
	}
	ci = cur.s.back;
	if ((first && (ci != VTE_DEF_BG)) || (old->s.back != ci)) {
		if (ci < VTE_COLOR_BRIGHT_OFFSET) {
			/* 40..47  base BG */
			g_string_append_printf(gs, "\033[%dm", 40 + ci);
		} else if (ci < VTE_COLOR_DIM_OFFSET) {
			/* 100..107  16-color BG */
			g_string_append_printf(gs, "\033[%dm",
				100 + ci - VTE_COLOR_BRIGHT_OFFSET);
		} else if (ci == VTE_DEF_BG) {
			/* default BG */
			g_string_append(gs, "\033[49m" );
		} else {
			/* 0..256  256-color BG  */
			g_string_append_printf(gs, "\033[48;5;%dm", ci);
		}
	}
	old->i = cur.i & STATE_MASK;
}

static void
dump(GtkWidget *window, gboolean new_file, GString *txt)
{
	int fd, count;
	const char *fname = g_object_get_data (G_OBJECT (window), "file");
	if (!fname) {
		g_print(txt->str);
		return;
	}

	char *cfname = g_object_get_data (G_OBJECT (window), "cfile");
	fd = GPOINTER_TO_INT(g_object_get_data (G_OBJECT (window), "fd"));
	count = GPOINTER_TO_INT(g_object_get_data (G_OBJECT (window), "count"));
	if (new_file && fd != -1) {
		if (close(fd) != 0) {
			if (close(fd) != 0) {
				g_printerr("Unable to close file '%s': %s\n",
					cfname == NULL ? "???" : cfname, g_strerror (errno));
				return;
			}
		}
		fd = -1;
	}

	if (fd == -1) {
		const char *fext = g_object_get_data (G_OBJECT (window), "fext");
		GString *file = g_string_new (fname);
		if (count != -1) {
			g_string_append_printf(file, "%d", count);
		}
		if (fext != NULL) {
			g_string_append_printf(file, ".%s", fext);
		}
		fd = g_open(file->str, O_WRONLY | O_CREAT | O_APPEND,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
		if (fd == -1 && errno != EINTR) {
			g_printerr("Unable to open file '%s' for writing: %s\n", file->str,
				g_strerror (errno));
			g_string_free(file, TRUE);
			return;
		}
		count++;
		g_object_set_data (G_OBJECT (window), "cfile", (gpointer) file->str);
		g_object_set_data (G_OBJECT (window), "fd", GINT_TO_POINTER (fd));
		g_object_set_data (G_OBJECT (window), "count", GINT_TO_POINTER (count));
		g_printerr("New file: %s\n", file->str);
		g_free(cfname);
		cfname = file->str;
		g_string_free(file, FALSE);
	}

	/* in contrast to stdout we make sure, that it always ends with an EOL */
	g_string_append_c(txt, '\n');
	gssize length = txt->len;
	char *contents = txt->str;
	while (length > 0) {
		gssize s;
		s = write (fd, contents, length);
		if (s < 0) {
			if (errno == EINTR)
				continue;
			g_printerr("Failed to write: %s\n", g_strerror (errno));
			break;
		}
		g_assert (s <= length);
		contents += s;
		length -= s;
	}
	g_printerr("Wrote %s\n", cfname);
}

static void
snapshot(GtkWidget *window, gboolean new_file, gboolean force)
{
	VteTerminal *terminal = VTE_TERMINAL(window); 
	VteScreen *screen = terminal->pvt->screen;
	VteRing *ring = screen->row_data;
	const VteCell *pcell = NULL;
	VteVisualPosition ss, se;
	VteIntCellAttr last_state, cur;
	PangoColor *palette, *orig_palette;
	PangoColor c0, c1;
	gboolean quirks = TRUE;

	glong col, row;
	GString *txt = g_string_new(NULL);

	/* determine the viewport */
	glong first_row = screen->scroll_delta;
	glong last_row = first_row + terminal->row_count - 1;
	glong last_col = terminal->column_count - 1;

	last_state.i = 0;
	last_state.s.fore = VTE_DEF_FG; last_state.s.back = VTE_DEF_BG;

	/*
	 * To have a consistent state and splittable results we initialize
	 * with/compare against initial defaults. Wrt. Escape codes see also:
	 * - http://invisible-island.net/xterm/ctlseqs/ctlseqs.html
	 * - http://invisible-island.net/xterm/xterm.faq.html#ctlseqs_ms
	 * - /usr/share/doc/xterm/ctlseqs.txt
	 */

	/* reset screen attrs */
	if (!quirks)
		g_string_append(txt, "\033[0m");

	/* report terminal size */
	row = last_row - first_row + 1;
	col = last_col + 1;
	g_string_append_printf(txt, "\n# Size: %dx%d => print '\\033[8;%d;%dt'\n\n",
		col, row, row, col);

	/* Check palette and adjust color settings if necessary: Since we have
	 * no history/marker, when a color has been changed, we simply assume,
	 * that is has been set before painting the viewport and got used in
	 * a consistent manner, i.e. not changed with the viewport again. */
	palette = terminal->pvt->palette;
	orig_palette = g_object_get_data(G_OBJECT(window), "palette");
	row = 0;
	for (col=0; col < VTE_PALETTE_SIZE; col++) {
		c0 = orig_palette[col];
		c1 = palette[col];
		if (c0.red == c1.red && c0.green == c1.green && c0.blue == c1.blue)
			continue;

		if (!row) {
			g_string_append(txt, "\033]4;");
			row++;
		}
		g_string_append_printf(txt, "%d;#%04x%04x%04x;",
			col, c1.red, c1.green, c1.blue);
	}
	if (row)
		g_string_overwrite(txt, txt->len -1, "\007");

	/* set video mode */
	screen->reverse_mode
		? g_string_append(txt, "\033[?5h")
		: g_string_append(txt, "\033[?5l");

	/* get selected area */
	ss = terminal->pvt->selection_start;
	se = terminal->pvt->selection_end;
	if (ss.row < 0 || se.row < 0 || !terminal->pvt->has_selection) {
		ss.row = last_row + 1;
	} else {
		if (ss.row < first_row) {
			ss.row = first_row;
			if (!terminal->pvt->selection_block_mode)
				ss.col = 0;
		}
		if (se.row > last_row) {
			se.row = last_row;
			if (!terminal->pvt->selection_block_mode)
				se.col = last_col;
		}
		if (ss.col < 0)
			ss.col = 0;
		if (se.col > last_col)
			se.col = last_col;
		if (ss.col == se.col
			&& (terminal->pvt->selection_block_mode || ss.row == se.row))
			ss.row = last_row + 1;
	}

	/* Quirks: less -R seems to reset the FG and BG colors to its defaults
	 * on NL but doesn't restore them at the next line (SOL). So we always do
	 * that fix on the SOL. With cat (who wonders) stuff gets displayed
	 * correctly, however, if the terminal is wider than the original, it may
	 * look not so good, i.e. BG color does not stop on NL but continues to
	 * the last column. So we do the same as less, reset at NL and than it
	 * shouldn't matter, whether one does a 'less -R' or 'cat' ... */

	/* print viewport, similar to vte_terminal_get_text() */
	col = 0;
	for (row = first_row; row < last_row + 1; row++, col = 0) {
		const VteRowData *row_data = NULL;

		if (G_LIKELY (_vte_ring_contains (ring, row))) {
			if (G_LIKELY (row >= ring->writable)) {
				row_data = &ring->array[row & ring->mask];
			} else {
				/* should not happen, because terminal updates are blocked
				 * and we explicitly ask for the viewport portion, only. */
				g_printerr("Unable to get data for row %d\n", row);
				if (force) {
					g_printerr("Termshot will be incomplete!\n");
				} else {
					g_printerr(
"Try to let your running program refresh the screen or press the SHIFT key\n" \
"as well to force recording an incomplete termshot.\nTermshot skipped!\n");
					g_string_free(txt, TRUE);
					return;
				}
			}
		}

		pcell = NULL;
		if (row_data != NULL) {
			while ((pcell = _vte_row_data_get (row_data, col))) {
				if (!pcell->attr.fragment) {
					cur.s = pcell->attr;
					if (row < ss.row || row > se.row) {
						cur.s.selected = FALSE;
					} else if (terminal->pvt->selection_block_mode
						|| (ss.row == se.row))
					{
						cur.s.selected = (col >= ss.col) && (col <= se.col);
					} else if (row == ss.row) {
						cur.s.selected = (col >= ss.col);
					} else if (row == se.row) {
						cur.s.selected = (col <= se.col);
					} else {
						cur.s.selected = TRUE;
					}
					adjust_state(&last_state, cur, quirks && col == 0, txt);
					if (pcell->c == 0) {
						g_string_append_c (txt, ' ');
					} else {
						if (G_UNLIKELY (pcell->c >= VTE_UNISTR_START)) {
							/* moooh - why is _vte_unistr_append_to_string()
							   not public */
							append_unistr(terminal, row, col,
								pcell->attr.columns, txt);
						} else {
							g_string_append_unichar(txt, (gunichar) pcell->c);
						}
					}
				}
				if ((row == last_row) && (col >= last_col)) {
					break;
				}
				col++;
			}
		}
		/* We do not strip off trailing WS, because this is an capture tool
		 * and should return the stuff AS IS.
		 * Furthermore we always append an EOL, no matter, whether the line
		 * is soft-wrappable or not for the same reason. */
		if (quirks)
			g_string_append(txt, "\033[0m");
		g_string_append_c(txt, '\n');
	}

	/* avoid wrapping for the last row */
	if (first_row < last_row + 1)
		g_string_truncate(txt, txt->len -1);

	dump(window, new_file, txt);
	g_string_free(txt, TRUE);
}

static int
button_pressed(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	/* this blocks an update of the terminal until the function returns */
	switch (event->button) {
	case 3:
		snapshot(widget, (event->state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK,
			(event->state & GDK_SHIFT_MASK) == GDK_SHIFT_MASK);
		break;
	default:
		break;
	}
	return FALSE;
}

static void
char_size_changed(GtkWidget *widget, guint width, guint height, gpointer data)
{
	VteTerminal *terminal;
	GtkWindow *window;
	GdkGeometry geometry;
	GtkBorder *inner_border;
		
	g_assert(GTK_IS_WINDOW(data));
	g_assert(VTE_IS_TERMINAL(widget));
		
	terminal = VTE_TERMINAL(widget);
	window = GTK_WINDOW(data);
	if (!gtk_widget_get_realized (GTK_WIDGET (window)))
		return;

	gtk_widget_style_get (widget, "inner-border", &inner_border, NULL);
	geometry.width_inc = width;
	geometry.height_inc = height;
	geometry.base_width =
		inner_border ? (inner_border->left + inner_border->right) : 0;
	geometry.base_height =
		inner_border ? (inner_border->top + inner_border->bottom) : 0;
	geometry.min_width = geometry.base_width + width * 2;
	geometry.min_height = geometry.base_height + height * 2;
	gtk_border_free (inner_border);

	gtk_window_set_geometry_hints(window, widget, &geometry,
		GDK_HINT_RESIZE_INC | GDK_HINT_BASE_SIZE | GDK_HINT_MIN_SIZE);
}


int
main(int argc, char **argv)
{
	GdkScreen *screen;
#if GTK_CHECK_VERSION (2, 90, 8)
	GdkVisual *visual;
#else
	GdkColormap *colormap;
#endif
	GtkWidget *window,*widget,*hbox = NULL, *scrollbar, *scrolled_window = NULL;
	VteTerminal *terminal;

	char *env_add[] = { NULL };
	const char *background = NULL;
	const char *encoding = NULL;
	gboolean audible = TRUE;
	char *geometry = NULL;
	gint lines = 250;
	const char *message = "Launching interactive shell...\r\n";
	const char *font = NULL;
	const char *termcap = NULL;
	const char *command = NULL;
	char *file = NULL;
	const char *fext = NULL;
	char *pty_flags_string = NULL;
	char *sb_policy_string = NULL;
	GdkColor fore, back;

	const GOptionEntry options[]={
		{
			"audible", 'a', G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, &audible,
			"Use visible, instead of audible, terminal bell",
			NULL
		},
		{
			"command", 'c', 0, G_OPTION_ARG_STRING, &command,
			"Execute the given command in the terminal. Default: $SHELL", "cmd"
		},
		{
			"encoding", 'e', 0, G_OPTION_ARG_STRING, &encoding,
			"Set the charset encoding to use. Default: UTF-8", "charset"
		},
		{
			"font", 'f', 0, G_OPTION_ARG_STRING, &font,
			"Specify a font to use (FontNameList[ StyleList][ Size]).  FontNameList is a comma separated list of font families, StyleList a whitespace separated list of one or more words for style, variant, weight, stretch, or gravity and Size a number (pt) optionally followed by px (pixels).",
			"FontSpec"
		},
		{
			"geometry", 'g', 0, G_OPTION_ARG_STRING, &geometry,
			"Set the size (in characters) and optionally position (e.g. 100x31).", "Geometry"
		},
		{
			"sb-lines", 'n', 0, G_OPTION_ARG_INT, &lines,
			"Specify the number of scrollback-lines. Default: 250", "Number"
		},
		{
			"out", 'o', 0, G_OPTION_ARG_STRING, &file,
			"Append terminal contents on each mouse button 3 (right) click to the given file. If one presses the Ctrl button at the same time, a new file gets created If not specified, the content gets printed to the standard output. Info/Warning messages are printed to standard error.", "File"
		},
		{
			"sb-policy", 'P', 0, G_OPTION_ARG_STRING, &sb_policy_string,
			"Set the policy for the vertical scroolbar (always|auto|never). Default:always)",
			"Policy"
		},
		{
			"pty-flags", 'p', 0, G_OPTION_ARG_STRING, &pty_flags_string,
			"A set of the following PTY flags: default|no-utmp|no-wtmp|no-lastlog|no-helper|no-fallback", "flags"
		},
		{
			"termcap", 't', 0, G_OPTION_ARG_STRING, &termcap,
			"Specify the terminal emulation to use. Default: xterm", "Term"
		},
		{ NULL }
	};
	GOptionContext *context;
	GError *error = NULL;
	GtkPolicyType scrollbar_policy = GTK_POLICY_ALWAYS;
	VtePtyFlags pty_flags = VTE_PTY_DEFAULT;

	context = g_option_context_new (" - VTE capture");
	g_option_context_add_main_entries (context, options, NULL);
	g_option_context_add_group (context, gtk_get_option_group (TRUE));
	g_option_context_parse (context, &argc, &argv, &error);
	g_option_context_free (context);
	if (error != NULL) {
		g_printerr ("Failed to parse command line arguments: %s\n",
			error->message);
		g_error_free (error);
		return 1;
	}

	if (sb_policy_string) {
		GEnumClass *enum_klass =
			(GEnumClass*)g_type_class_ref(GTK_TYPE_POLICY_TYPE);
		const GEnumValue *enum_value =
			g_enum_get_value_by_nick(enum_klass, sb_policy_string);
		if (enum_value) {
			scrollbar_policy = enum_value->value;
		} else {
			g_warning("Unknown enum '%s'\n", sb_policy_string);
		}
		g_type_class_unref(enum_klass);
		g_free(sb_policy_string);
	}
	if (pty_flags_string) {
		char **flags = flags = g_strsplit_set(pty_flags_string, ",|", -1);

		if (flags != NULL) {
			guint i;
			GFlagsClass *flags_klass =
				(GFlagsClass*)g_type_class_ref(VTE_TYPE_PTY_FLAGS);
			for (i = 0; flags[i] != NULL; ++i) {
				const GFlagsValue *flags_value;

				flags_value = g_flags_get_value_by_nick(flags_klass, flags[i]);
				if (flags_value)
					pty_flags |= flags_value->value;
				else
					g_warning("Unknown flag '%s'\n", flags[i]);
			}
			g_type_class_unref(flags_klass);
		}
		g_free(pty_flags_string);
	}

	back.red = back.green = 0xffff; back.blue = 0xe800;
	fore.red = fore.green = fore.blue = 0x0000;

	/* Create a window to hold the scrolling shell, and hook its
	 * delete event to the quit function */
	gdk_window_set_debug_updates(FALSE);
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_container_set_resize_mode(GTK_CONTAINER(window), GTK_RESIZE_IMMEDIATE);

	/* Set ARGB colormap */
	screen = gtk_widget_get_screen (window);
#if GTK_CHECK_VERSION (2, 90, 8)
	visual = gdk_screen_get_rgba_visual(screen);
	if (visual)
		gtk_widget_set_visual(GTK_WIDGET(window), visual);
#else
	colormap = gdk_screen_get_rgba_colormap (screen);
	if (colormap)
	    gtk_widget_set_colormap(window, colormap);
#endif

	/* Create a box to hold everything */
	hbox = gtk_hbox_new(0, FALSE);
	gtk_container_add(GTK_CONTAINER(window), hbox);

	/* Create the terminal widget and add it to the scrolling shell */
	widget = vte_terminal_new();
	terminal = VTE_TERMINAL (widget);

	gtk_box_pack_start(GTK_BOX(hbox), widget, TRUE, TRUE, 0);

	/* make the geometry params work */
	char_size_changed(widget, 0, 0, window);
	g_signal_connect(widget, "char-size-changed",
		G_CALLBACK(char_size_changed), window);

	/* Connect to the "button-press" event */
	g_signal_connect(widget, "button-press-event",
		G_CALLBACK(button_pressed), widget);

	/* Create the scrollbar for the widget */
	scrollbar = gtk_vscrollbar_new(terminal->adjustment);
	gtk_box_pack_start(GTK_BOX(hbox), scrollbar, FALSE, FALSE, 0);

	/* Set some defaults */
	vte_terminal_set_audible_bell(terminal, audible);
	vte_terminal_set_visible_bell(terminal, !audible);
	vte_terminal_set_cursor_blink_mode(terminal, VTE_CURSOR_BLINK_SYSTEM);
	vte_terminal_set_scroll_background(terminal, FALSE);
	vte_terminal_set_scroll_on_output(terminal, FALSE);
	vte_terminal_set_scroll_on_keystroke(terminal, TRUE);
	vte_terminal_set_scrollback_lines(terminal, lines);
	vte_terminal_set_mouse_autohide(terminal, TRUE);
	vte_terminal_set_colors(terminal, &fore, &back, NULL, 0);
	vte_terminal_set_opacity(terminal, 0xffff);
	if (termcap != NULL) {
		vte_terminal_set_emulation(terminal, termcap);
	}
	vte_terminal_set_cursor_shape(terminal, VTE_CURSOR_SHAPE_BLOCK);
	vte_terminal_set_encoding(terminal, encoding == NULL ? "UTF-8" : encoding);

	/* Set the default font */
	vte_terminal_set_font_from_string_full(terminal, font,
		VTE_ANTI_ALIAS_USE_DEFAULT);

	GError *err = NULL;
	char **command_argv = NULL;
	int command_argc;
	GPid pid = -1;

	if (command == NULL || *command == '\0')
		command = g_getenv ("SHELL");

	if (command == NULL || *command == '\0')
		command = "/bin/sh";

	if (!g_shell_parse_argv(command, &command_argc, &command_argv, &err)
		|| !vte_terminal_fork_command_full(terminal, pty_flags, NULL,
			command_argv, env_add, G_SPAWN_SEARCH_PATH, NULL, NULL,
			&pid, &err))
	{
		g_warning("Failed to fork: %s\n", err->message);
		g_error_free(err);
	} else {
		g_printerr("Capture Terminal PID %d\n", pid);
	}

	g_strfreev(command_argv);
	
	/* try to split off file extension, so that we do not need to do it again
	 * and again */
	gint open = -1;
	if (file) {
		GFile *f = g_file_new_for_path(file);
		char *basename = g_file_get_basename(f);
		char *dot = strrchr(basename, '.');
		int elen = dot == NULL ? 0 : strlen(dot);
		if (elen > 1) {
			g_free(basename);
			basename = file;
			dot = file + (strlen(file) - elen);
			fext = strdup(dot+1);
			*dot = '\0';
			file = strdup(file);
			*dot = '.';
			g_object_set_data (G_OBJECT (widget), "file", (gpointer) file);
			g_object_set_data (G_OBJECT (widget), "fext", (gpointer) fext);
		} else {
			g_object_set_data (G_OBJECT (widget), "file", (gpointer) file);
			g_object_set_data (G_OBJECT (widget), "fext", NULL);
		}
		g_free(basename);
		g_object_unref(f);
	} else {
		g_object_set_data (G_OBJECT (widget), "file", NULL);
		g_object_set_data (G_OBJECT (widget), "fext", NULL);
	}
	g_object_set_data (G_OBJECT (widget), "fd", GINT_TO_POINTER (open));
	g_object_set_data (G_OBJECT (widget), "count", GINT_TO_POINTER (open));

	/* Go for it! */
	g_signal_connect(widget, "child-exited", G_CALLBACK(child_exited), window);
	g_signal_connect(window, "delete-event", G_CALLBACK(delete_event), widget);

	g_object_add_weak_pointer(G_OBJECT(widget), (gpointer*)&widget);
	g_object_add_weak_pointer(G_OBJECT(window), (gpointer*)&window);

	gtk_widget_realize(widget);
	if (geometry) {
		if (!gtk_window_parse_geometry (GTK_WINDOW(window), geometry)) {
			g_printerr("Could not parse --geometry argument");
		}
	}
#if GTK_CHECK_VERSION(2, 91, 0)
	else {
		/* As of GTK+ 2.91.0, the default size of a window comes from its min
		 * size not its natural size, so we need to set the right default size
		 * explicitly */
		gtk_window_set_default_geometry (GTK_WINDOW (window),
			vte_terminal_get_column_count (terminal),
			vte_terminal_get_row_count (terminal));
	}
#endif

	gtk_widget_show_all(window);

	/* Make a copy of the initial color palette to be able to detect changes
	 * See also: vte_terminal_set_colors() */
	g_assert(sizeof(terminal->pvt->palette)/sizeof(PangoColor)
		== VTE_PALETTE_SIZE);
	gpointer p =
		g_memdup(terminal->pvt->palette, sizeof(terminal->pvt->palette));
	g_object_set_data (G_OBJECT (widget), "palette", p);

	gtk_main();

	g_assert(widget == NULL);
	g_assert(window == NULL);

	return 0;
}
/*
# vim: ts=4 sw=4 filetype=c
*/
